from typing import Any, Tuple

import gymnasium as gym
import numpy as np

from .agent import Agent


class RandomAgent(Agent):
    """An agent that takes random actions in the environment.

    :param name: The name of the agent.
    :type name: str
    :param action_space: The action space of the environment.
    :type action_space: gym.Space
    :param num_actions: The number of actions the agent can take.
    :type num_actions: int
    :param seed: The seed for the random number generator.
    :type seed: int, optional
    """

    def __init__(
        self, name: str, action_space: gym.Space, num_actions: int, seed: int = None
    ) -> None:
        super().__init__(name, seed)

        self._action_space = action_space
        self._num_actions = num_actions

    def learn(self, **kwargs: dict[str, Any]) -> None:
        pass

    def predict(
        self, observation: dict[str, Any], **kwargs: dict[str, Any]
    ) -> Tuple[np.ndarray, Any]:
        ants = observation["ants"]
        max_ants = self._action_space.shape[0]

        # Create a mask for each ant
        # Each mask is a binary array of shape (num_actions,)
        # where 1 indicates that the action is valid for the ant
        masks = []
        for i in range(max_ants):
            # If the ant is alive, then it can take any action
            if ants[i]:
                masks.append(np.ones(self._num_actions, dtype=np.int8))
            # If no ant is present (or dead), then it can only take the no-op action
            else:
                mask = np.zeros(self._num_actions, dtype=np.int8)
                mask[-1] = 1
                masks.append(mask)

        # Sample a random action for each ant using the mask
        # `sample()` expects a tuple
        actions = self._action_space.sample(mask=tuple(masks))

        return actions, None

    def save(self, path: str, **kwargs: dict[str, Any]) -> None:
        pass

    def load(self, path: str, **kwargs: dict[str, Any]) -> None:
        pass



from abc import ABC, abstractmethod
from typing import Any, Tuple

import numpy as np


class Agent(ABC):
    """Defines the base class for all agents.

    :param name: The name of the agent
    :type name: str
    :param seed: The seed for the random number generator, defaults to None
    :type seed: int, optional
    """

    def __init__(self, name: str, seed: int = None) -> None:
        super().__init__()

        self._name = name
        self._seed = seed

    @abstractmethod
    def learn(self, **kwargs: dict[str, Any]) -> None:
        """Runs a learning procedure to train the agent"""
        raise NotImplementedError

    @abstractmethod
    def predict(
        self, observation: dict[str, Any], **kwargs: dict[str, Any]
    ) -> Tuple[np.ndarray, Any]:
        """Predicts the action(s) to take given the input observation

        :param observation: The input observation to predict for
        :type observation: dict[str, Any]
        :return: The action(s) to take, and any other extra information (optional)
        :rtype: Tuple[np.ndarray, Any]
        """
        raise NotImplementedError

    @abstractmethod
    def save(self, path: str, **kwargs: dict[str, Any]) -> None:
        """Saves the agent's parameters and related artifacts to the given path

        :param path: The path to save to
        :type path: str
        """
        raise NotImplementedError

    @abstractmethod
    def load(self, path: str, **kwargs: dict[str, Any]) -> None:
        """Loads the agent's artifacts from the given path

        :param path: The path to load from
        :type path: str
        """
        raise NotImplementedError



from typing import List, Optional

class Action:
    """A class representing an action that an ant can take.

    Attributes:
        row (int): The row of the location of the ant.
        col (int): The column of the location of the ant.
        direction (Direction): The direction of the movement.
    """

    row: int
    """The row of the location of the ant."""
    col: int
    """The column of the location of the ant."""
    direction: Direction
    """The direction of the movement."""

class Ant:
    """A class representing an ant.

    Attributes:
        id (str): The unique identifier for the ant.
        row (int): The row of the location of the ant.
        col (int): The column of the location of the ant.
        player (int): The player that owns the ant.
        alive (bool): Whether the ant is alive.
        field_of_vision (List[Entity]): The field of vision of the ant as a list of entities the ant can see.
    """

    id: str
    """The unique identifier for the ant."""
    row: int
    """The row of the location of the ant."""
    col: int
    """The column of the location of the ant."""
    player: int
    """The player that owns the ant."""
    alive: bool
    """Whether the ant is alive."""
    field_of_vision: List[Entity]
    """The field of vision of the ant as a list of entities the ant can see."""

class Direction:
    """An enum representing a direction.

    Attributes:
        North (str): The North direction.
        East (str): The East direction.
        South (str): The South direction.
        West (str): The West direction.
    """

    North: str
    """The North direction."""
    East: str
    """The East direction."""
    South: str
    """The South direction."""
    West: str
    """The West direction."""

class Entity:
    """A class representing an entity.

    Attributes:
        name (str): The name of the entity. "Ant", "Food", "Hill" or "Water".
        row (int): The row of the location of the entity.
        col (int): The column of the location of the entity.
        player (int): The player that owns the entity.
        alive (bool): Whether the entity is alive, only applicable to ants.
    """

    name: str
    """The name of the entity. "Ant", "Food", "Hill" or "Water"."""
    row: int
    """The row of the location of the entity."""
    col: int
    """The column of the location of the entity."""
    player: int
    """The player that owns the entity."""
    alive: bool
    """Whether the entity is alive, only applicable to ants."""

class FinishedReason:
    """An enum representing the reason the game finished.

    Attributes:
        LoneSurvivor (str): The game ended because there was only one player left.
        RankStabilized (str): The game ended because the rank stabilized, i.e. no player can surpass the current leader anymore.
        TooMuchFood (str): The game ended because food was not being consumed and it reached 90% or more of the map.
        TurnLimitReached (str): The game ended because the maximum number of turns was reached.
    """

    LoneSurvivor: str
    """The game ended because there was only one player left."""
    RankStabilized: str
    """The game ended because the rank stabilized, i.e. no player can surpass the current leader anymore."""
    TooMuchFood: str
    """The game ended because food was not being consumed and it reached 90% or more of the map."""
    TurnLimitReached: str
    """The game ended because the maximum number of turns was reached."""

class Game:
    """A class representing the Ants game. Main entry point for the environment.

    :param map_contents: The contents of the map file.
    :type map_contents: str
    :param fov_radius2: The squared radius of the field of vision of the ants.
    :type fov_radius2: int
    :param attack_radius2: The squared radius of the attack range of the ants.
    :type attack_radius2: int
    :param food_radius2: The squared radius of the range around ants to harvest food.
    :type food_radius2: int
    :param food_rate: The amount of food to spawn *per player* on each round.
    :type food_rate: int
    :param max_turns: The maximum number of turns for the Ants game.
    :type max_turns: int
    :param max_colony_size: The maximum number of live ants a player can have at any time.
    :type max_colony_size: int
    :param seed: The seed for the random number generator.
    :type seed: int
    :param replay_filename: The filename to save the replay of the game to. If `None`, no replay will be saved.
    :type replay_filename: str, optional
    """

    def __init__(
        self,
        map_contents: str,
        fov_radius2: int,
        attack_radius2: int,
        food_radius2: int,
        food_rate: int,
        max_turns: int,
        max_colony_size: int,
        seed: int,
        replay_filename: Optional[str],
    ) -> None: ...
    def width(self) -> int:
        """Returns the width of the map.

        :return: The width of the map.
        :rtype: int
        """

    def height(self) -> int:
        """Returns the height of the map.

        :return: The height of the map.
        :rtype: int
        """

    def players(self) -> int:
        """Returns the number of players in the game.

        :return: The number of players in the game.
        :rtype: int
        """

    def start(self) -> GameState:
        """Starts the game.

        Must be called once before updating the game state.

        :return: The initial game state.
        :rtype: GameState
        """

    def update(self, actions: List[Action]) -> GameState:
        """Updates the game state with the given actions.

        :param actions: The actions to take for each ant.
        :type actions: List[Action]
        :return: The updated game state.
        :rtype: GameState
        """

    def draw(self) -> None:
        """Draws the current state of the game."""

class GameState:
    """A class representing the state of the game.

    Attributes:
        turn (int): The current turn number.
        scores (List[int]): The scores for each player.
        ants (List[List[Ant]]): The list of ants for each player.
        hive (List[int]): The list of the number of ants in each player's hive.
        turn_stats (List[TurnStats]): The list of turn stats for each player.
        finished (bool): Whether the game has finished.
        finished_reason (Optional[FinishedReason]): The reason the game finished. Only present if the game has finished.
        winner (Optional[int]): The player that won the game. Only present if the game has finished and there is a winner.
    """

    turn: int
    """The current turn number."""
    scores: List[int]
    """The scores for each player."""
    ants: List[List[Ant]]
    """The list of ants for each player."""
    hive: List[int]
    """The list of the number of ants in each player's hive."""
    turn_stats: List[TurnStats]
    """The list of turn stats for each player."""
    finished: bool
    """Whether the game has finished."""
    finished_reason: Optional[FinishedReason]
    """The reason the game finished. Only present if the game has finished."""
    winner: Optional[int]
    """The player that won the game. Only present if the game has finished and there is a winner."""

class TurnStats:
    """A class representing the stats for a turn.

    Attributes:
        turn (int): The turn number.
        food_harvested (int): The amount of food harvested by the player.
        ants_spawned (int): The number of ants spawned by the player.
        ants_killed (int): The number of ants killed by the player.
        hills_razed (int): The number of hills razed by the player.
        ants_lost (int): The number of ants lost by the player.
        hills_lost (int): The number of hills lost by the player.
    """

    turn: int
    """The turn number."""
    food_harvested: int
    """The amount of food harvested by the player."""
    ants_spawned: int
    """The number of ants spawned by the player."""
    ants_killed: int
    """The number of ants killed by the player."""
    hills_razed: int
    """The number of hills razed by the player."""
    ants_lost: int
    """The number of ants lost by the player."""
    hills_lost: int
    """The number of hills lost by the player."""



from .agents.agent import Agent
from .agents.random import RandomAgent
from .ants_ai import Action, Ant, Direction, Entity, FinishedReason
from .env import AntsEnv
from .visualizer import Visualizer

__all__ = [
    "Action",
    "Agent",
    "Ant",
    "AntsEnv",
    "Entity",
    "Direction",
    "FinishedReason",
    "RandomAgent",
    "Visualizer",
]



from typing import Any, List, Tuple

import gymnasium as gym
import numpy as np

from .agents import Agent, RandomAgent
from .ants_ai import Action, Direction, FinishedReason, Game

type ActType = np.ndarray
type InfoType = dict[str, Any]
type ObsType = dict[str, Any]

RL_AGENT_PLAYER_INDEX = 0
FOOD_HARVESTED_MULTIPLIER = 1.0
ANTS_SPAWNED_MULTIPLIER = 1.0
ANTS_KILLED_MULTIPLIER = 2.0
ANTS_LOST_MULTIPLIER = 2.0
HILLS_RAZED_MULTIPLIER = 10.0
HILL_LOST_MULTIPLIER = 10.0
END_GAME_BONUS = 100.0
LIVING_PENALTY = 0.1


class AntsEnv(gym.Env):
    """The Ants environment.

    :param map_file: The path to the map file.
    :type map_file: str
    :param fov_radius2: The radius **squared** of the field of vision for each ant, defaults to 77.
    :type fov_radius2: int, optional
    :param attack_radius2: The radius **squared** of the attack range for each ant, defaults to 5.
    :type attack_radius2: int, optional
    :param food_radius2: The radius **squared** of the range around ants to harvest food, defaults to 1
    :type food_radius2: int, optional
    :param food_rate: The amount of food to spawn *per player* on each round, defaults to 5.
    :type food_rate: int, optional
    :param max_turns: The maximum number of turns for the Ants game, defaults to 1500.
    :type max_turns: int, optional
    :param max_colony_size: The maximum number of live ants a player can have at any time, defaults to 500.
    :type max_colony_size: int, optional
    :param other_agents: The other agents to use in the game. If `None`, agents that act randomly will be created for each other player.
    :type other_agents: List[Agent], optional
    :param seed: The seed for the random number generator, defaults to 0.
    :type seed: int, optional
    :param replay_filename: The filename to save the replay of the game to. If `None`, no replay will be saved, defaults to `None`.
    :type replay_filename: _type_, optional
    :param render_mode: The render mode to use. If `None`, no render is computed. Defaults to `None`.
    :type render_mode: str, optional
    """

    metadata = {"render_modes": ["human"]}

    def __init__(
        self,
        map_file: str,
        fov_radius2: int = 77,
        attack_radius2: int = 5,
        food_radius2: int = 1,
        food_rate: int = 5,
        max_turns: int = 1500,
        max_colony_size: int = 500,
        other_agents: List[Agent] = None,
        seed: int = 0,
        replay_filename: str = None,
        render_mode: str = None,
    ):
        super().__init__()

        with open(map_file) as f:
            map_contents = f.read()

        self.game = Game(
            map_contents,
            fov_radius2,
            attack_radius2,
            food_radius2,
            food_rate,
            max_turns,
            max_colony_size,
            seed,
            replay_filename,
        )

        self.channels = 11
        self.observation_space = gym.spaces.Dict(
            {
                # The observation space is a 2D grid representing a partially observable map.
                # The first channel represents the visibility "mask", i.e. `1` means that the entity is visible and `0` means that it is not.
                # The remaining channels represent the entities in the game, as follows:
                # 1: Live colony (i.e. ants of the player)
                # 2: Dead colony (i.e. ants of the player)
                # 3: Enemy colonies (i.e. ants of the enemies)
                # 4: Dead enemy colonies (i.e. ants of the enemies)
                # 5: Food
                # 6: Hills (i.e. the hills of the player)
                # 7: Razed hills (i.e. the hills of the player)
                # 8: Enemy hills (i.e. the hills of the enemies)
                # 9: Razed enemy hills (i.e. the hills of the enemies)
                # 10: Water
                # The observation is channel-first.
                "map": gym.spaces.Box(
                    low=0,
                    high=1,
                    shape=(self.channels, self.game.width(), self.game.height()),
                    dtype=int,
                ),
                # This extra space is used to represent the ants of the player.
                # It's an array where each element is a binary value indicating whether the ant is alive or not.
                "ants": gym.spaces.MultiBinary(max_colony_size),
            },
            seed=seed,
        )
        self.num_actions = 5
        # The action space is a list of actions for each ant.
        # The possible actions are: N, E, S, W, Stay
        self.action_space = gym.spaces.MultiDiscrete(
            [self.num_actions] * max_colony_size, seed=seed
        )

        self._max_colony_size = max_colony_size
        # Tracks the latest game state
        self._game_state = None
        # Tracks the index in the action space of each ant per player
        self._ant_id_to_index = {player: {} for player in range(self.game.players())}
        # Tracks the next index available to use for the next ant per player
        self._next_index_per_player = {
            player: 0 for player in range(self.game.players())
        }
        # Track reusable indices from dead ants per player
        self._free_indices_per_player = {
            player: [] for player in range(self.game.players())
        }
        self._action_to_direction = {
            0: Direction.North,
            1: Direction.East,
            2: Direction.South,
            3: Direction.West,
            4: "Stay",
        }

        self._validate_other_agents(other_agents)
        self._other_agents = (
            other_agents
            if isinstance(other_agents, list)
            else [
                RandomAgent(
                    name=f"Player {i + 1}",
                    action_space=self.action_space,
                    num_actions=self.num_actions,
                    seed=seed + i,
                )
                for i in range(self.game.players() - 1)
            ]
        )

        self._render_mode = render_mode

    def reset(self, seed=None, options=None) -> Tuple[ObsType, InfoType]:
        """Resets the environment.

        :return: The initial observation and info.

                 - The observation is a 2D grid representing a partially observable map and the vector of ants.
                 - The info is a dictionary with the keys `turn`, `scores`, `done_reason`, `winner` and `game_state`.
                         - `turn` is the current turn of the game.
                         - `scores` is a list of scores for each player.
                         - `done_reason` is the reason why the game is done, if it is done.
                         - `winner` is the index of the winner, if there is one.
                         - `game_state` is the raw game state (see `GameState`), useful for heuristic-based agents and not suitable for RL agents.
        :rtype: Tuple[ObsType, InfoType]
        """
        super().reset(seed=seed, options=options)

        self._ant_id_to_index = {player: {} for player in range(self.game.players())}
        self._next_index_per_player = {
            player: 0 for player in range(self.game.players())
        }
        self._free_indices_per_player = {
            player: [] for player in range(self.game.players())
        }

        game_state = self.game.start()

        self._game_state = game_state
        self._update_index_mapping()

        return self._get_obs(player=RL_AGENT_PLAYER_INDEX), self._get_info()

    def step(self, action: ActType) -> Tuple[ObsType, float, bool, bool, InfoType]:
        """Takes a step in the environment.

        :param action: The action to take. The action is an array of actions for each ant.
        :type actions: ActType
        :return: The observation, reward, whether the game is done, whether the game was truncated and extra info (see `reset()`).
        :rtype: Tuple[ObsType, float, bool, bool, InfoType]
        """
        # Map the RL agent's action to the game actions
        game_actions: List[Action] = []
        for player in range(self.game.players()):
            if player == RL_AGENT_PLAYER_INDEX:
                raw_action = action
            else:
                # Get the action from the other agents
                # We subtract 1 because `other_agents` is a 0-indexed list of the _other_ players
                # and the first other player is player 1 (i.e. index 0).
                raw_action, _ = self._other_agents[player - 1].predict(
                    self._get_obs(player=player)
                )

            for ant in self._game_state.ants[player]:
                index = self._ant_id_to_index[player][ant.id]
                _action = raw_action[index]
                direction = self._action_to_direction[_action]

                if direction == "Stay":
                    # `Stay` means the ant does not move (i.e. doesn't take any action)
                    continue

                game_actions.append(Action(ant.row, ant.col, direction))

        game_state = self.game.update(game_actions)

        self._game_state = game_state
        self._update_index_mapping()

        return (
            self._get_obs(player=RL_AGENT_PLAYER_INDEX),
            self._get_reward(),
            game_state.finished,
            # Truncated is always False, since the game is not truncated
            False,
            self._get_info(),
        )

    def render(self) -> None:
        """Renders the current state of the environment to the console.

        It is recommended to use `replay_filename` in the constructor to save the replay to a file, and use the visualizer to view the replay instead of using this method.
        """
        if self._render_mode is None or self._render_mode != "human":
            return

        self.game.draw()

    def _get_obs(self, player: int) -> ObsType:
        ants = self._game_state.ants[player]
        indices = self._ant_id_to_index[player]
        minimap = np.zeros(
            (self.channels, self.game.width(), self.game.height()), dtype=int
        )
        ants_mask = np.zeros(self._max_colony_size, dtype=int)

        # 0: Visibility mask
        # 1: Live colony (i.e. ants of the player)
        # 2: Dead colony (i.e. ants of the player)
        # 3: Enemy colonies (i.e. ants of the enemies)
        # 4: Dead enemy colonies (i.e. ants of the enemies)
        # 5: Food
        # 6: Hills (i.e. the hills of the player)
        # 7: Razed hills (i.e. the hills of the player)
        # 8: Enemy hills (i.e. the hills of the enemies)
        # 9: Razed enemy hills (i.e. the hills of the enemies)
        # 10: Water
        for ant in ants:
            # Add the ant to the mask
            index = indices[ant.id]
            ants_mask[index] = 1

            # Add the ant to the minimap
            minimap[0, ant.col, ant.row] = 1
            minimap[1, ant.col, ant.row] = 1

            # Add the field of vision of the ant to the minimap
            for entity in ant.field_of_vision:
                minimap[0, entity.col, entity.row] = 1

                if entity.name == "Ant":
                    colony = 1 if entity.player == ant.player else 3
                    alive = 1 - int(entity.alive)
                    minimap[colony + alive, entity.col, entity.row] = 1
                elif entity.name == "Food":
                    minimap[5, entity.col, entity.row] = 1
                elif entity.name == "Hill":
                    colony = 6 if entity.player == ant.player else 8
                    alive = 1 - int(entity.alive)
                    minimap[colony + alive, entity.col, entity.row] = 1
                elif entity.name == "Water":
                    minimap[10, entity.col, entity.row] = 1

        return {
            "map": minimap,
            "ants": ants_mask,
        }

    def _get_info(self) -> InfoType:
        return {
            "turn": self._game_state.turn,
            "scores": self._game_state.scores,
            "done_reason": str(self._game_state.finished_reason),
            "winner": self._game_state.winner,
            "game_state": self._game_state,
        }

    def _get_reward(self) -> float:
        reward = 0.0
        stats = self._game_state.turn_stats[RL_AGENT_PLAYER_INDEX]

        reward += stats.food_harvested * FOOD_HARVESTED_MULTIPLIER
        reward += stats.ants_spawned * ANTS_SPAWNED_MULTIPLIER
        reward += stats.ants_killed * ANTS_KILLED_MULTIPLIER
        reward += stats.hills_razed * HILLS_RAZED_MULTIPLIER
        reward -= stats.ants_lost * ANTS_LOST_MULTIPLIER
        reward -= stats.hills_lost * HILL_LOST_MULTIPLIER

        reward -= LIVING_PENALTY

        # Add bonus/penalty at the end of the game
        if self._game_state.finished:
            if (
                self._game_state.finished_reason
                in (FinishedReason.LoneSurvivor, FinishedReason.RankStabilized)
                and self._game_state.winner == RL_AGENT_PLAYER_INDEX
            ):
                reward += END_GAME_BONUS
            else:
                reward -= END_GAME_BONUS

        return float(reward)

    def _update_index_mapping(self) -> None:
        for player, ants in enumerate(self._game_state.ants):
            self._free_dead_ants_indices(player)

            for ant in ants:
                if ant.id in self._ant_id_to_index[player]:
                    continue

                if self._free_indices_per_player[player]:
                    index = self._free_indices_per_player[player].pop()
                else:
                    index = self._next_index_per_player[player]

                    if index >= self._max_colony_size:
                        raise ValueError(
                            f"Player {player} has too many ants ({index}/{self._max_colony_size}). This is a bug!"
                        )
                    self._next_index_per_player[player] += 1

                self._ant_id_to_index[player][ant.id] = index

    def _free_dead_ants_indices(self, player: int) -> None:
        dead_ant_ids = set(self._ant_id_to_index[player]) - set(
            [ant.id for ant in self._game_state.ants[player]]
        )
        for dead_ant_id in dead_ant_ids:
            index = self._ant_id_to_index[player][dead_ant_id]
            self._free_indices_per_player[player].append(index)
            del self._ant_id_to_index[player][dead_ant_id]

    def _validate_other_agents(self, other_agents: List[Agent]) -> None:
        if other_agents is None:
            return

        if len(other_agents) != self.game.players() - 1:
            raise ValueError(
                f"Number of agents ({len(other_agents)}) doesn't match number of other players ({self.game.players() - 1})."
            )



import importlib.resources
import json
import os
import re
from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass
from enum import Enum
from typing import Any, Optional, Self

# Hide the annoying pygame support prompt
os.environ["PYGAME_HIDE_SUPPORT_PROMPT"] = "1"
import pygame

PLAYER_COLORS = {
    0: (212, 17, 50),
    1: (232, 110, 48),
    2: (230, 179, 25),
    3: (246, 246, 85),
    4: (166, 229, 153),
    5: (46, 184, 126),
    6: (48, 140, 232),
    7: (99, 23, 207),
    8: (214, 92, 214),
    9: (207, 175, 183),
}
UPDATE_EPSILON = 0.001
REFERENCE_WINDOW_SIZE = 800
REFERENCE_FONT_SIZE = 20
PLAYERS_LABEL = "Players: "
LIVE_ANTS_LABEL = "Live ants: "
SCORES_LABEL = "Scores: "
HIVE_LABEL = "Hive: "
INFO_LABEL_OFFSET = 10


class TurnPhase(Enum):
    """An enumeration representing the different phases of a turn."""

    Spawn = "Spawn"
    Move = "Move"
    Attack = "Attack"
    Remove = "Remove"

    def all() -> list[Self]:
        """Returns a list of all the turn phases."""
        return [phase for phase in TurnPhase]


@dataclass
class Event:
    """A class representing an event that occurred in the game.

    Attributes:
        event_type (str): The type of event. "Spawn", "Remove", "Move" or "Attack".
        entity (str): The entity associated with the event. "Ant", "Food" or "Hill".
        entity_id (str): The entity ID associated with the
        player (int): The player that owns the entity.
        location (tuple[int]): The location of the entity as a tuple of (row, col).
        destination (tuple[int]): The destination of the entity as a tuple of (row, col). Only used for "Move" and "Attack" events.
    """

    event_type: str
    """The type of event. "Spawn", "Remove", "Move" or "Attack"."""
    entity: str
    """The entity associated with the event. "Ant", "Food" or "Hill"."""
    entity_id: str
    """The entity ID associated with the event."""
    player: int
    """The player that owns the entity."""
    location: tuple[int]
    """The location of the entity as a tuple of (row, col)."""
    destination: tuple[int]
    """The destination of the entity as a tuple of (row, col). Only used for "Move" and "Attack" events."""


@dataclass
class Turn:
    """A class representing a turn in the game.

    Attributes:
        turn_number (int): The number of the turn.
        ants (list[int]): The number of live ants per player.
        hive (list[int]): The number of ants in the hive per player.
        scores (list[int]): The scores of the players.
        events (dict[TurnPhase, list[Event]]): A dictionary mapping turn phases to the events that occurred in that phase.
    """

    turn_number: int
    """The number of the turn."""
    ants: list[int]
    """The number of live ants per player."""
    hive: list[int]
    """The number of ants in the hive per player."""
    scores: list[int]
    """The scores of the players."""
    events: dict[TurnPhase, list[Event]]
    """A dictionary mapping turn phases to the events that occurred in that phase."""

    @classmethod
    def from_json(cls, dict: dict[str, Any]) -> Self:
        """Creates a `Turn` object from a JSON dictionary.

        :param dict: The JSON dictionary to create the `Turn` object from.
        :type dict: dict[str, Any]
        :return: The `Turn` object created from the JSON dictionary.
        :rtype: Self
        """
        # Group events by phase
        events = defaultdict(list)
        for event in dict["events"]:
            phase = TurnPhase(event["event_type"])
            events[phase].append(Event(**event))

        return cls(
            dict["turn"],
            dict["ants"],
            dict["hive"],
            dict["scores"],
            events,
        )


@dataclass
class Map:
    """A class representing the map of the game.

    Attributes:
        width (int): The width of the map.
        height (int): The height of the map.
        contents (str): The contents of the map.
    """

    width: int
    """The width of the map."""
    height: int
    """The height of the map."""
    contents: str
    """The contents of the map."""


@dataclass
class Replay:
    """A class representing a replay of a full game.

    Attributes:
        players (int): The number of players in the game.
        map (Map): The map of the game.
        turns (list[Turn]): The list of turns in the game.
        finished_reason (str): The reason the game finished.
        winner (Optional[int]): The player that won the game, if any.
    """

    players: int
    """The number of players in the
    game."""
    map: Map
    """The map of the game."""
    turns: list[Turn]
    """The list of turns in the game."""
    finished_reason: str
    """The reason the game finished."""
    winner: Optional[int]
    """The player that won the game, if any."""

    @classmethod
    def from_json(cls, dict: dict[str, Any]) -> Self:
        """Creates a `Replay` object from a JSON dictionary.

        :param dict: The JSON dictionary to create the `Replay` object from.
        :type dict: dict[str, Any]
        :return: The `Replay` object created from the JSON dictionary.
        :rtype: Self
        """
        turns = list(map(lambda turn: Turn.from_json(turn), dict["turns"]))
        _map = Map(**dict["map"])

        return cls(
            dict["players"],
            _map,
            turns,
            cls._parse_finished_reason(dict["finished_reason"]),
            dict["winner"],
        )

    @classmethod
    def _parse_finished_reason(cls, finished_reason: str) -> str:
        if finished_reason == "LoneSurvivor":
            return "Lone Survivor"
        elif finished_reason == "RankStabilized":
            return "Rank Stabilized"
        elif finished_reason == "TooMuchFood":
            return "Too Much Food"
        elif finished_reason == "TurnLimitReached":
            return "Max Turns Reached"
        else:
            return finished_reason


@dataclass
class Entity(ABC):
    """An abstract class representing an entity in the game.

    Attributes:
        id: (str): The ID of the entity.
        location (tuple[int]): The location of the entity as a tuple of (row, col).
        target_location (tuple[int]): The target location of the entity as a tuple of (row, col).
        scale (int): The scale of the entity on the screen.
        size (int): The size of the entity.
        target_size (int): The target size of the entity.
        alive (bool): Whether the entity is alive or not.
    """

    id: str = ""
    """The ID of the entity."""
    location: tuple[int] = (-1, -1)
    """The location of the entity as a tuple of (row, col)."""
    target_location: tuple[int] = (-1, -1)
    """The target location of the entity as a tuple of (row, col)."""
    scale: int = 0
    """The scale of the entity on the screen."""
    size: int = 0
    """The size of the entity."""
    target_size: int = 0
    """The target size of the entity."""
    alive: bool = False
    """Whether the entity is alive or not."""

    @abstractmethod
    def draw(self, screen: pygame.Surface) -> None:
        """Draws the entity.

        :param screen: The screen to draw the entity on.
        :type screen: pygame.Surface
        """
        raise NotImplementedError

    def update(self, phase: int, dt: float) -> None:
        """Updates the entity.

        :param phase: The current phase.
        :type phase: int
        :param dt: The time since the last update.
        :type dt: float
        """
        update_t = dt - phase
        phase = TurnPhase.all()[phase]

        if phase in [TurnPhase.Spawn, TurnPhase.Remove]:
            # Update the size
            if self.size == self.target_size:
                return

            size_update = (self.target_size - self.size) * update_t
            self.size += size_update

            # Cap the size to 0 when shrinking and the target size when growing
            self.size = (
                max(0, self.size)
                if self.target_size == 0
                else min(self.target_size, self.size)
            )

            # If the size is within UPDATE_EPSILON of the target size, snap to the target size
            if abs(self.size - self.target_size) <= UPDATE_EPSILON:
                self.size = self.target_size
        elif phase == TurnPhase.Move:
            self.location = self._update_location(
                self.location, self.target_location, update_t
            )

    def _update_location(
        self,
        location: tuple[int],
        target_location: tuple[int],
        update_t: float,
    ) -> tuple[int]:
        if location == target_location:
            return location

        current_row, current_col = location
        target_row, target_col = target_location

        # Update the appropriate location
        if target_row != current_row:
            row_diff = target_row - current_row
            row_update = row_diff * update_t
            new_row = current_row + row_update

            # Cap the new row to the target row based on the direction
            new_row = (
                min(target_row, new_row) if row_diff > 0 else max(target_row, new_row)
            )

            # If the new row is within UPDATE_EPSILON of the target row, snap to the target row
            if abs(new_row - target_row) <= UPDATE_EPSILON:
                new_row = target_row

            current_row = new_row

        if target_col != current_col:
            col_diff = target_col - current_col
            col_update = col_diff * update_t
            new_col = current_col + col_update

            # Cap the new col to the target col based on the direction
            new_col = (
                min(target_col, new_col) if col_diff > 0 else max(target_col, new_col)
            )

            # If the new col is within UPDATE_MOVE_EPSILON of the target col, snap to the target col
            if abs(new_col - target_col) <= UPDATE_EPSILON:
                new_col = target_col

            current_col = new_col

        return current_row, current_col


@dataclass
class Ant(Entity):
    """A class representing an ant in the game.

    Attributes:
        player (int): The player that owns the ant.
    """

    player: int = -1
    """The player that owns the ant."""

    def draw(self, screen: pygame.Surface) -> None:
        row, col = self.location
        color = PLAYER_COLORS[self.player]
        radius = self.scale // 2
        center = (col * self.scale + radius, row * self.scale + radius)
        pygame.draw.circle(screen, color, center, self.size)


@dataclass
class Food(Entity):
    """A class representing food in the game."""

    def draw(self, screen: pygame.Surface) -> None:
        row, col = self.location
        offset = (self.scale - self.size) // 2
        rect = (
            col * self.scale + offset,
            row * self.scale + offset,
            self.size,
            self.size,
        )
        pygame.draw.rect(screen, (153, 145, 102), rect)


@dataclass
class Hill(Entity):
    """A class representing a hill in the game.

    Attributes:
        player (int): The player that owns the hill.
        sprites (tuple[pygame.Surface]): The sprites for the hill (alive and razed).
    """

    player: int = -1
    """The player that owns the hill."""
    sprites: tuple[pygame.Surface] = (None, None)
    """The sprites for the hill (alive and razed)."""

    def draw(self, screen: pygame.Surface) -> None:
        row, col = self.location
        sprite = self.sprites[0] if self.alive else self.sprites[1]
        sprite = pygame.transform.scale(sprite, (self.scale, self.scale))

        # Draw an outline of the player's color on the sprite to indicate ownership of the hill
        # Only draw the outline if the hill is alive
        if self.alive:
            color = PLAYER_COLORS[self.player]
            overlay = pygame.Surface((self.scale, self.scale), pygame.SRCALPHA)
            center = (self.scale // 2, self.scale // 2)
            radius = self.scale // 4
            pygame.draw.circle(overlay, color, center, radius, width=3)
            sprite.blit(overlay, (0, 0))

        screen.blit(sprite, (col * self.scale, row * self.scale))


@dataclass
class Water(Entity):
    """A class representing water in the game.

    Attributes:
        sprite: (pygame.Surface): The sprite to use for the water.
    """

    sprite: pygame.Surface = None
    """The sprite to use for the water."""

    def draw(self, screen: pygame.Surface) -> None:
        row, col = self.location
        sprite = pygame.transform.scale(self.sprite, (self.scale, self.scale))
        screen.blit(sprite, (col * self.scale, row * self.scale))


@dataclass
class Attack(Entity):
    """A class representing an attack in the game.

    Attributes:
        current_target_location (tuple[int]): The current target location of the attack that will be updated towards the target location.
    """

    current_target_location: tuple[int] = (-1, -1)
    """The current target location of the attack that will be updated towards the target location."""

    def draw(self, screen: pygame.Surface) -> None:
        row, col = self.location
        dest_row, dest_col = self.current_target_location
        # Offset to center the line
        offset = self.scale // 2
        pygame.draw.line(
            screen,
            (255, 255, 255),
            (col * self.scale + offset, row * self.scale + offset),
            (dest_col * self.scale + offset, dest_row * self.scale + offset),
            2,
        )

    def update(self, phase, dt):
        # Update the current _target_ location towards the target location
        update_t = dt - phase
        self.current_target_location = self._update_location(
            self.current_target_location, self.target_location, update_t
        )


class Visualizer:
    """A class for visualizing a replay of a full Ants game.

    :param replay_filename: The filename of the replay to visualize.
    :type replay_filename: str
    :param scale: The scale factor for the map when visualizing, defaults to 10.
    :type scale: int
    :param show_grid: Whether to show the grid lines on the map, defaults to False.
    :type show_grid: bool
    """

    def __init__(
        self,
        replay_filename: str,
        scale: int = 10,
        show_grid: bool = False,
    ) -> None:
        pygame.init()
        pygame.display.set_caption("Ants Replay Visualizer")

        self._hill_sprites = self._load_hill_sprites()
        self._water_sprite = self._load_water_sprite()

        self._replay = self._load_replay(replay_filename)
        self._width = self._replay.map.width
        self._height = self._replay.map.height
        self._scale = scale

        self._water: list[Water] = []
        self._hills: dict[tuple[int], Hill] = {}
        self._food: dict[tuple[int], Food] = {}
        self._ants: dict[str, Ant] = {}
        self._attacks: list[Attack] = []
        self._parse_map()

        self._land_color = (120, 89, 58)
        self._game_size = (self._width * self._scale, self._height * self._scale)

        with importlib.resources.path(
            "ants_ai.assets", "RobotoMono-Regular.ttf"
        ) as font_path:
            scale_factor = min(
                self._game_size[0] / REFERENCE_WINDOW_SIZE,
                self._game_size[1] / REFERENCE_WINDOW_SIZE,
            )
            font_size = int(REFERENCE_FONT_SIZE * scale_factor)
            self._font = pygame.font.Font(str(font_path), font_size)

        self._info_size = (
            self._game_size[0],
            # Make sure that the info panel is big enough to fit all the labels
            int(
                self._font.size(PLAYERS_LABEL)[1]
                + self._font.size(LIVE_ANTS_LABEL)[1]
                + self._font.size(SCORES_LABEL)[1]
                + self._font.size(HIVE_LABEL)[1]
                + INFO_LABEL_OFFSET
            ),
        )
        self._window_size = (
            self._game_size[0],
            self._game_size[1] + self._info_size[1],
        )

        self._screen = pygame.display.set_mode(self._window_size)
        self._game_screen = pygame.Surface(self._game_size)
        self._info_screen = pygame.Surface(self._info_size)

        self._clock = pygame.time.Clock()
        # Simulation time
        self._time = 0.0
        # Per-turn time
        self._turn_time = 0.0

        self._show_grid = show_grid
        self._turn_phases = TurnPhase.all()
        self._replayed: dict[int, dict[TurnPhase, bool]] = {
            turn: {phase: False for phase in self._turn_phases}
            for turn in range(len(self._replay.turns))
        }

    def run(self) -> None:
        """Runs the visualizer."""
        running = True
        finished = False
        turn = 0
        prev_turn = 0

        while running:
            if finished:
                running = not self._should_quit()
                continue

            dt = self._clock.tick(60) / 1000.0
            self._time += dt

            # Track the per-turn time
            self._turn_time = (self._time * len(self._turn_phases)) % len(
                self._turn_phases
            )

            turn = int(self._time)

            if turn != prev_turn:
                # Remove dead entities and attack lines at the start of each turn
                self._remove_dead_entities()
                self._attacks.clear()
                prev_turn = turn

                # Check if the simulation has finished
                finished = turn >= len(self._replay.turns)
                if finished:
                    self._draw_endgame(
                        self._replay.finished_reason, self._replay.winner
                    )
                    # Re-draw the game surface to show the endgame message
                    self._screen.blit(self._game_screen, (0, self._info_size[1]))
                    pygame.display.flip()
                    continue

            phase_index = int(self._turn_time)
            phase = self._turn_phases[phase_index]

            if not self._replayed[turn][phase]:
                events = self._replay.turns[turn].events[phase]
                self._do_replay(events)
                self._replayed[turn][phase] = True

            self._update_map(phase_index)

            self._draw_info(turn, self._replay.turns[-1].turn_number)
            self._draw_grid()
            self._draw_map()

            # Draw the full screen as a combination of the game and info surfaces
            self._screen.blit(self._info_screen, (0, 0))
            self._screen.blit(self._game_screen, (0, self._info_size[1]))

            pygame.display.flip()

            running = not self._should_quit()

    def _should_quit(self) -> bool:
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                return True
        return False

    def _draw_info(self, turn: int, last_turn: int) -> None:
        self._info_screen.fill((0, 0, 0))
        width, _ = self._info_size

        # Draw number of players
        players_text_width, players_text_height = self._draw_text(
            f"{PLAYERS_LABEL}{self._replay.players}",
            color=(255, 255, 255),
            location=(INFO_LABEL_OFFSET, INFO_LABEL_OFFSET),
            surface=self._info_screen,
        )

        # Draw the turn number
        turn_text = f"Turn: {turn} / {last_turn}"
        turn_text_width, _ = self._font.size(turn_text)
        # Center the turn number in the middle of the info surface after the players text
        area_width = width - players_text_width - INFO_LABEL_OFFSET
        turn_text_location = (
            players_text_width
            + INFO_LABEL_OFFSET
            + area_width // 2
            - turn_text_width // 2,
            INFO_LABEL_OFFSET,
        )
        self._draw_text(
            turn_text,
            color=(255, 255, 255),
            location=turn_text_location,
            surface=self._info_screen,
        )

        live_ants_text_width = self._font.size(LIVE_ANTS_LABEL)[0]
        scores_text_width = self._font.size(SCORES_LABEL)[0]
        hive_text_width = self._font.size(HIVE_LABEL)[0]

        # Draw the player scores as a colored bar proportional to the score
        self._draw_players_bar(
            label=SCORES_LABEL,
            label_location=(INFO_LABEL_OFFSET, INFO_LABEL_OFFSET + players_text_height),
            data=self._replay.turns[turn].scores,
            # Align it with the "Live ants" bar
            offset_x=INFO_LABEL_OFFSET + live_ants_text_width - scores_text_width,
        )

        # Draw the number of live ants as a colored bar proportional to the number of ants
        self._draw_players_bar(
            label=LIVE_ANTS_LABEL,
            label_location=(
                INFO_LABEL_OFFSET,
                INFO_LABEL_OFFSET + players_text_height * 2,
            ),
            data=self._replay.turns[turn].ants,
            offset_x=INFO_LABEL_OFFSET,
        )

        # Draw the number of ants in the hive as a colored bar proportional to the number of ants
        self._draw_players_bar(
            label=HIVE_LABEL,
            label_location=(
                INFO_LABEL_OFFSET,
                INFO_LABEL_OFFSET + players_text_height * 3,
            ),
            data=self._replay.turns[turn].hive,
            # Align it with the "Live ants" bar
            offset_x=INFO_LABEL_OFFSET + live_ants_text_width - hive_text_width,
        )

    def _draw_players_bar(
        self,
        label: str,
        label_location: tuple[int],
        data: list[int],
        offset_x: int,
    ) -> None:
        label_width, label_height = self._draw_text(
            label,
            color=(255, 255, 255),
            location=label_location,
            surface=self._info_screen,
        )

        bar_width = self._info_size[0] - offset_x - label_width
        bar_height = label_height

        total = sum(data)
        current_offset_x = offset_x + label_width
        offset_y = label_location[1]

        # If there is no data to draw, draw a "0" in the center
        if total == 0:
            self._draw_text(
                "0",
                color=(255, 255, 255),
                location=(current_offset_x + bar_width // 2, offset_y),
                surface=self._info_screen,
            )
            return

        for i, value in enumerate(data):
            # Draw a bar proportional to the value
            portion = value / total if total > 0 else 0
            color = PLAYER_COLORS[i]
            width = int(portion * bar_width)
            rect = (
                current_offset_x,
                offset_y,
                width,
                bar_height,
            )
            pygame.draw.rect(self._info_screen, color, rect)

            # Draw the value text in the center of the bar
            value_text = f"{value}"
            self._draw_text(
                value_text,
                color=(0, 0, 0),
                location=(current_offset_x + width // 2, offset_y),
                surface=self._info_screen,
            )

            current_offset_x += width

        # Draw a border around the bar
        pygame.draw.rect(
            self._info_screen,
            (255, 255, 255),
            (offset_x + label_width, offset_y, bar_width, bar_height),
            2,
        )

    def _draw_grid(self) -> None:
        if not self._show_grid:
            return

        for row in range(self._height):
            for col in range(self._width):
                rect = (
                    col * self._scale,
                    row * self._scale,
                    self._scale,
                    self._scale,
                )
                pygame.draw.rect(self._game_screen, (0, 0, 0), rect, 1)

    def _draw_map(self) -> None:
        self._game_screen.fill(self._land_color)
        for entity in [
            *self._water,
            *self._hills.values(),
            *self._food.values(),
            *self._ants.values(),
            *self._attacks,
        ]:
            entity.draw(self._game_screen)

    def _draw_endgame(self, finished_reason: str, winner: Optional[int]) -> None:
        _, height = self._draw_finished_reason(finished_reason)
        self._draw_winner(winner, offset_y=height)

    def _draw_finished_reason(self, finished_reason: str) -> tuple[int]:
        return self._draw_text_center(
            f"Game Over: {finished_reason}", color=(255, 255, 255)
        )

    def _draw_winner(self, winner: Optional[int], offset_y: int) -> None:
        if winner is None:
            return

        self._draw_text_center(
            text=f"Player {winner} won!",
            color=PLAYER_COLORS[winner],
            offset=(0, offset_y),
        )

    def _draw_text_center(
        self, text: str, color: tuple[int], offset: tuple[int] = (0, 0)
    ) -> tuple[int]:
        text_width, text_height = self._font.size(text)
        location = (
            offset[0] + self._game_size[0] // 2 - text_width // 2,
            offset[1] + self._game_size[1] // 2 - text_height // 2,
        )

        return self._draw_text(
            text, color=color, location=location, surface=self._game_screen
        )

    def _draw_text(
        self,
        text: str,
        color: tuple[int],
        location: tuple[int],
        surface: pygame.Surface,
    ) -> tuple[int]:
        surface.blit(
            self._font.render(text, True, color),
            location,
        )

        return self._font.size(text)

    def _update_map(self, phase: int) -> None:
        for entity in [
            *self._food.values(),
            *self._ants.values(),
            *self._attacks,
        ]:
            entity.update(phase, self._turn_time)

    def _do_replay(self, events: list[Event]) -> None:
        for event in events:
            if event.event_type == "Spawn":
                self._replay_spawn(event)
            elif event.event_type == "Remove":
                self._replay_remove(event)
            elif event.event_type == "Move":
                self._replay_move(event)
            elif event.event_type == "Attack":
                self._replay_attack(event)
            else:
                raise RuntimeError(
                    f"Invalid event type '{event.event_type}' in event {event}."
                )

    def _replay_spawn(self, event: Event) -> None:
        location = tuple(event.location)

        if event.entity == "Ant":
            self._ants[event.entity_id] = self._spawn_ant(
                event.entity_id, location, event.player
            )
        elif event.entity == "Food":
            self._food[location] = self._spawn_food(location)
        else:
            raise RuntimeError(
                f"Invalid 'Spawn' event for entity '{event.entity}': {event}."
            )

    def _replay_remove(self, event: Event) -> None:
        if event.entity == "Ant":
            self._ants[event.entity_id].target_size = 0
            self._ants[event.entity_id].alive = False
        elif event.entity == "Food":
            location = tuple(event.location)
            self._food[location].target_size = 0
            self._food[location].alive = False
        elif event.entity == "Hill":
            location = tuple(event.location)
            # When hills are removed they are "razed", not removed from the map
            self._hills[location].alive = False
        else:
            raise RuntimeError(
                f"Invalid 'Remove' event for entity '{event.entity}': {event}."
            )

    def _replay_move(self, event: Event) -> None:
        to = tuple(event.destination)
        ant = self._ants.get(event.entity_id)

        if not ant:
            raise RuntimeError(
                f"No ant found with id {event.entity_id} to move in event: {event}."
            )

        # Move the ant to its new location
        ant.target_location = to

    def _replay_attack(self, event: Event) -> None:
        self._attacks.append(
            Attack(
                location=tuple(event.location),
                target_location=tuple(event.destination),
                current_target_location=tuple(event.location),
                scale=self._scale,
            )
        )

    def _remove_dead_entities(self) -> None:
        # Remove dead ants
        for ant_id, ant in list(self._ants.items()):
            if not ant.alive:
                del self._ants[ant_id]

        # Remove consumed/destroyed food
        for location, food in list(self._food.items()):
            if not food.alive:
                del self._food[location]

    def _load_hill_sprites(self) -> tuple[pygame.Surface]:
        with importlib.resources.path("ants_ai.assets", "hill.png") as img_path:
            spritesheet = pygame.image.load(str(img_path))
            w, h = spritesheet.get_size()
            # We have 2 states (alive and razed) stacked vertically
            sprite_height = h // 2
            # Extract the "alive" anthill (top half)
            alive = spritesheet.subsurface((0, 0, w, sprite_height))
            # Extract the "razed" anthill (bottom half)
            razed = spritesheet.subsurface((0, sprite_height, w, sprite_height))

        return alive, razed

    def _load_water_sprite(self) -> pygame.Surface:
        with importlib.resources.path("ants_ai.assets", "water.png") as img_path:
            return pygame.image.load(str(img_path))

    def _load_replay(self, replay_filename: str) -> Replay:
        with open(replay_filename, "r") as file:
            return Replay.from_json(json.load(file))

    def _parse_map(self) -> None:
        regex = re.compile(r"m (.*)")

        for row, line in enumerate(regex.finditer(self._replay.map.contents)):
            for col, char in enumerate(line.group(1).strip()):
                # Ignore land
                if char == ".":
                    continue

                location = (row, col)

                # Max 10 players
                if "0" <= char <= "9":
                    player = int(char)
                    sprites = [
                        self._hill_sprites[0].copy(),
                        self._hill_sprites[1].copy(),
                    ]
                    self._hills[location] = Hill(
                        id=f"Hill(p={player},loc=({location}))",
                        location=location,
                        scale=self._scale,
                        alive=True,
                        player=player,
                        sprites=sprites,
                    )
                elif char == "*":
                    self._food[location] = self._spawn_food(location)
                elif char == "%":
                    self._water.append(
                        Water(
                            location=location,
                            scale=self._scale,
                            alive=True,
                            sprite=self._water_sprite,
                        )
                    )
                else:
                    raise ValueError(
                        f"Invalid entity in map with character value: {char}"
                    )

    def _spawn_ant(self, id: str, location: tuple[int], player: int) -> Ant:
        return Ant(
            id,
            location,
            target_location=location,
            scale=self._scale,
            size=0,
            target_size=self._scale // 5,
            alive=True,
            player=player,
        )

    def _spawn_food(self, location: tuple[int]) -> Food:
        return Food(
            id=f"Food(loc=({location}))",
            location=location,
            target_location=location,
            scale=self._scale,
            size=0,
            target_size=self._scale // 3,
            alive=True,
        )



import time
from pathlib import Path

import numpy as np

from ants_ai import AntsEnv, RandomAgent, Visualizer


def main() -> None:
    map_file = Path(__file__).parent / "maps" / "tutorial.map"

    env = AntsEnv(map_file, replay_filename="/tmp/random_agent_replay.json")
    p1 = RandomAgent("RL Agent", env.action_space, env.num_actions, seed=42)

    start = time.time()
    done = False
    obs, _ = env.reset(seed=42)
    rewards = []
    while not done:
        action, _ = p1.predict(obs)
        obs, reward, done, _, info = env.step(action)
        rewards.append(reward)

    print("Game finished")
    print(f"Avg. Reward: {np.average(rewards)}")
    print(f"Reason: {info['done_reason']}")
    print(f"Winner: {info['winner']}")
    print(f"Game took {time.time() - start} seconds")

    Visualizer("/tmp/random_agent_replay.json", scale=20, show_grid=True).run()


if __name__ == "__main__":
    main()



import argparse
from pathlib import Path
from typing import Any, Optional, Tuple

import gymnasium as gym
import numpy as np
import torch
import torch.nn as nn
from stable_baselines3 import PPO
from stable_baselines3.common.distributions import Categorical, Distribution
from stable_baselines3.common.policies import ActorCriticPolicy
from stable_baselines3.common.torch_layers import BaseFeaturesExtractor

from ants_ai import Agent, AntsEnv, Visualizer


class PartiallyObservableMapFeatureExtractor(BaseFeaturesExtractor):
    def __init__(self, observation_space: gym.spaces.Dict) -> None:
        map_shape = observation_space["map"].shape
        features_size = 128

        super().__init__(observation_space, features_dim=features_size)

        self.cnn = nn.Sequential(
            nn.Conv2d(map_shape[0], 16, kernel_size=3, padding=1),
            nn.ReLU(),
            nn.Flatten(),
        )

        # Compute shape by doing one forward pass
        # We need to do this to avoid computing the shape manually
        with torch.no_grad():
            n_flatten = self.cnn(torch.zeros(1, *map_shape)).shape[1]

        self.linear = nn.Sequential(
            nn.Linear(n_flatten, features_size),
            nn.ReLU(),
        )

    def forward(self, observations: dict[str, Any]) -> torch.Tensor:
        return self.linear(self.cnn(observations["map"]))


class MaskedPolicy(ActorCriticPolicy):
    def __init__(
        self, *args, max_colony_size: int = 25, num_actions: int = 5, **kwargs
    ):
        super().__init__(
            *args,
            features_extractor_class=PartiallyObservableMapFeatureExtractor,
            normalize_images=False,
            **kwargs,
        )

        self.max_colony_size = max_colony_size
        self.num_actions = num_actions
        self.action_net = nn.Linear(
            self.features_dim, self.max_colony_size * self.num_actions
        )
        self.value_net = nn.Linear(self.features_dim, 1)

    def forward(self, obs: dict[str, Any], deterministic: bool = False) -> torch.Tensor:
        features = self.extract_features(obs)
        masked_logits = self._mask(obs, features)
        dist = Categorical(logits=masked_logits)

        actions = (
            dist.sample() if not deterministic else torch.argmax(masked_logits, dim=-1)
        )
        log_prob = dist.log_prob(actions).sum(dim=1)

        return actions, self.value_net(features), log_prob

    def evaluate_actions(
        self, obs: dict[str, Any], actions: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor, Optional[torch.Tensor]]:
        features = self.extract_features(obs)
        masked_logits = self._mask(obs, features)
        dist = Categorical(logits=masked_logits)
        values = self.value_net(features)
        log_prob = dist.log_prob(actions).sum(dim=1)
        entropy = dist.entropy().sum(dim=1)

        return values, log_prob, entropy

    def predict_values(self, obs: dict[str, Any]) -> torch.Tensor:
        return self.value_net(self.extract_features(obs))

    def get_distribution(self, obs: dict[str, Any]) -> Distribution:
        features = self.extract_features(obs)
        masked_logits = self._mask(obs, features)
        dist = Categorical(logits=masked_logits)

        return dist

    def _predict(
        self, observation: dict[str, Any], deterministic: bool = False
    ) -> torch.Tensor:
        actions, _, _ = self.forward(observation, deterministic)

        return actions

    def _mask(
        self, observation: dict[str, Any], features: torch.Tensor
    ) -> torch.Tensor:
        # Shape: (batch, max_colony_size * num_actions)
        logits = self.action_net(features)
        logits = logits.view(-1, self.max_colony_size, self.num_actions)

        # Mask padded ants
        # Shape: (batch, max_colony_size)
        mask = observation["ants"].bool()

        # Set logits to -inf for invalid ants
        masked_logits = logits.clone()
        masked_logits[~mask] = float("-inf")

        # Set logits to 0 for the no-op action (last action)
        batch_idx, ant_idx = torch.where(~mask)
        masked_logits[batch_idx, ant_idx, self.num_actions - 1] = 0.0

        return masked_logits


class RLAgent(Agent):
    def __init__(
        self,
        name: str,
        env: gym.Env,
        max_colony_size: int,
        num_actions: int,
        seed: int = None,
    ) -> None:
        super().__init__(name, seed)

        self._env = env
        self._model = PPO(
            MaskedPolicy,
            env,
            policy_kwargs=dict(
                max_colony_size=max_colony_size,
                num_actions=num_actions,
            ),
            seed=seed,
            tensorboard_log="/tmp/tensorboard/",
        )

    def learn(self, **kwargs: dict[str, Any]) -> None:
        self._model.learn(**kwargs, tb_log_name=self._name)

    def predict(
        self, observation: dict[str, Any], **kwargs: dict[str, Any]
    ) -> Tuple[np.ndarray, Any]:
        return self._model.predict(observation, deterministic=True)

    def save(self, path: str, **kwargs: dict[str, Any]) -> None:
        self._model.save(path)

    def load(self, path: str, **kwargs: dict[str, Any]) -> None:
        self._model = PPO.load(path, env=self._env)


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--train", action="store_true", help="Train the agent (default: False)"
    )
    parser.add_argument(
        "--eval", action="store_true", help="Evaluate the agent (default: False)"
    )
    args = parser.parse_args()

    if args.train and args.eval:
        raise ValueError("Cannot train and evaluate at the same time.")

    map_file = Path(__file__).parent / "maps" / "tutorial.map"
    max_colony_size = 25
    save_path = "/tmp/ants_tutorial_ppo"
    env = AntsEnv(
        map_file,
        food_rate=15,
        max_turns=500,
        max_colony_size=max_colony_size,
        replay_filename="/tmp/tutorial_replay.json" if args.eval else None,
        seed=42,
    )
    agent = RLAgent("RL Agent", env, max_colony_size, env.num_actions, seed=42)

    if args.train:
        _train(agent, save_path)
    elif args.eval:
        agent.load(save_path)
        _eval(env, agent)


def _train(agent: Agent, save_path: str) -> None:
    print(f"Training agent {agent._name}")

    agent.learn(total_timesteps=10_000, progress_bar=True)
    agent.save(save_path)


def _eval(env: gym.Env, agent: Agent) -> None:
    print(f"Evaluating agent {agent._name}!")

    obs, _ = env.reset(seed=42)
    done = False
    rewards = []
    while not done:
        action, _ = agent.predict(obs, deterministic=True)
        obs, reward, done, _, info = env.step(action)
        rewards.append(reward)

    print("Game finished")
    print(f"Avg. Reward: {np.average(rewards)}")
    print(f"Reason: {info['done_reason']}")
    print(f"Winner: {info['winner']}")

    Visualizer("/tmp/tutorial_replay.json", scale=20, show_grid=True).run()


if __name__ == "__main__":
    main()
